def clamp(value: float, smallest: float = 0, largest: float = 1) -> float:
	''' Returns the value clamped between smallest and largest.
	I.e.: `clamp(10, 2, 8)` would return 8. '''
	return max(smallest, min(value, largest))

def snap(value: float, step: float = 1, offset: float = 0) -> float:
	''' Returns the value snapped to a scale of size step with an optional offset.
	I.e.: `snap(3.1, 2, 0)` would return 4. '''
	return round((value - offset) / step) * step + offset

def unique(seq: list, uniqueness_function) -> list:
	''' Returns a list in the same order with just the elements with a unique
	value on the uniqueness_function.
	I.e.: `unique([1, 5, 7, 9], lambda x: x % 3)` would return [1, 5, 9]. '''
	values = []
	keys = []
	for elem in seq:
		k = uniqueness_function(elem)
		if k not in keys:
			keys.append(k)
			values.append(elem)
	return values

def all_equal(seq: list) -> list:
	''' Returns true if every element in the sequence has the same value. '''
	return len(seq) < 2 or all(seq[0] == e for e in seq[1:])

def jaccard(seq1: list, seq2: list) -> list:
	''' Returns the Jaccard index of two sequences. '''
	set1, set2 = set(seq1), set(seq2)
	return len(set1.intersection(set2)) / len(set1.union(set2))

def levenshtein(seq1: list, seq2: list) -> list:
	''' Returns the Levenshtein distance of two sequences. '''
	from difflib import SequenceMatcher
	return SequenceMatcher(None, seq1, seq2).ratio()

def base_change(n: list, base_from: int, base_to: int) -> list:
	''' Changes the base of n represented as a list of integers.
	Example:  base_change([1, 1, 0, 1], 2, 10) = [1, 3]
	'''
	res = []
	n = sum(base_from ** (len(n) - i - 1) * v for i, v in enumerate(n))
	while n:
		n, digit = divmod(n, base_to)
		res.insert(0, digit)
	return res

def prime_list(n: int) -> list:
	''' Returns the list of prime numbers from 2 to n. '''
	from math import sqrt, ceil
	sieve = [True] * int(n / 2)
	for i in range(3, ceil(sqrt(n)), 2):
		if sieve[int(i / 2)]:
			sieve[int(i * i / 2)::i] = [False] * int((n - i * i - 1) / (2 * i) + 1)
	return [2] + [2 * i + 1 for i in range(1, int(n / 2)) if sieve[i]]

def is_prime(n: int) -> bool:
	''' Checks if a number is prime. '''
	from math import sqrt
	res = True
	if n == 2 or n == 3:
		res = True
	elif n < 2 or n % 2 == 0:
		res = False
	elif n < 9:
		res = True
	elif n % 3 == 0:
		res = False
	r = int(sqrt(n))
	f = 5
	while f <= r:
		if n % f == 0 or n % (f + 2) == 0:
			res = False
		else:
			f += 6
	return res

def fibonacci(n: int) -> int:
	''' Returns the n-th Fibonacci number. '''
	def _fib(n):
		if n == 0:
			return (0, 1)
		else:
			a, b = _fib(n // 2)
			c = a * (2 * b - a)
			d = b * b + a * a
			if n % 2 == 0:
				return (c, d)
			else:
				return (d, c + d)
	return _fib(n)[0]

def lcm(a: int, b: int) -> int:
	''' Least common multiple of two numbers. '''
	return a * b / gcd(a, b)

def gcd(a: int, b: int) -> int:
	''' Greatest common divisor of two numbers. '''
	if a < 0:
		a = -a
	if b < 0:
		b = -b
	if a == 0:
		return b
	while (b):
		a, b = b, a % b
	return a

def factor(n: int) -> list:
	''' Returns the factors of n and its exponents. '''
	f, factors, prime_gaps = 1, [], [2, 4, 2, 4, 6, 2, 6, 4]
	if n < 1:
		return []
	while True:
		for gap in ([1, 1, 2, 2, 4] if f < 11 else prime_gaps):
			f += gap
			if f * f > n:  # If f > sqrt(n)
				if n == 1:
					return factors
				else:
					return factors + [(n, 1)]
			if not n % f:
				e = 1
				n //= f
				while not n % f:
					n //= f
					e += 1
				factors.append((f, e))

def palindrome_list(k: int) -> list:
	''' Returns a list of every palindromic number with k digits. '''
	from itertools import product
	if k == 1:
		return [1, 2, 3, 4, 5, 6, 7, 8, 9]
	return [
		sum([
			n * (10**i)
			for i, n in enumerate(
				([x] + list(ys) + [z] + list(ys)[::-1] + [x])
				if k % 2 else ([x] + list(ys) + list(ys)[::-1] + [x])
			)
		])
		for x in range(1, 10)
		for ys in product(range(10), repeat=k // 2 - 1)
		for z in (range(10) if k % 2 else (None,))
	]

def phi(n: int) -> int:
	''' Returns the Euler's phi function of n. '''
	from functools import reduce
	from itertools import combinations
	res = n
	factors = [f[0] for f in factor(n)]
	multiplier = 1
	for fs in range(1, len(factors) + 1):
		multiplier *= -1
		for primes in combinations(factors, fs):
			val = reduce(lambda x, y: x * y, primes)
			if val <= n and n % val == 0:
				res += (n // val) * multiplier
	return res

def date_range(date_start, date_end, step=None):
	from datetime import timedelta, datetime
	assert isinstance(date_start, datetime) and isinstance(date_end, datetime), 'date_start and date_end must be datetime objects'
	assert isinstance(step, timedelta), 'step must be a timedelta object'
	if step is None:
		step = timedelta(days=1)
	current = date_start
	while current < date_end:
		yield current
		current += step


def human_bytes(size: float, decimal_places: int = 2) -> str:
	''' Returns a human readable file size from a number of bytes. '''
	for unit in ['', 'k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y']:
		if size < 1024: break
		size /= 1024
	return f'{size:.{decimal_places}f}{unit}B'

def human_seconds(seconds: float) -> str:
	''' Returns a human readable string from a number of seconds. '''
	from datetime import timedelta
	return str(timedelta(seconds=int(seconds))).zfill(8)

_human_date_measures = (
	('year', 365 * 24 * 3600),
	('month', 30 * 24 * 3600),
	('week', 7 * 24 * 3600),
	('day', 24 * 3600),
	('hour', 3600),
	('minute', 60),
	('second', 1)
)
def human_date(date) -> str:
	''' Return a date the a human-friendly format "1 month ago". '''
	from datetime import datetime
	if isinstance(date, str):
		res = datetime.strptime(date, '%Y-%m-%d, %H:%M:%S')
	else:
		res = date
	diff = (datetime.now() - res).total_seconds()
	if diff < 1:
		return 'just now'
	else:
		for name, amount in _human_date_measures:
			if diff > amount:
				diff = diff // amount
				return '%d %s%s ago' % (diff, name, 's' if diff > 1 else '')

def random_string(length: int, mask: list = None) -> str:
	''' Returns a random string. '''
	from random import choice
	if mask is None:
		from string import digits, ascii_letters
		mask = digits + ascii_letters
	return ''.join(choice(mask) for _ in range(length))

def print_matrix(matrix: list, rows: int = None, cols: int = None, elem_width: int = None, separator: str = ' ') -> str:
	for row in matrix[:rows]:
		for col in row[:cols]:
			print(str(col)[:elem_width], end=separator)
		print()

_safe_filename_regex = None
def safe_filename(filename: str) -> str:
	global _safe_filename_regex
	if _safe_filename_regex is None:
		from regex import compile
		_safe_filename_regex = compile(r'[/\\\*;\[\]\":=,<>]')
	return _safe_filename_regex.sub('', filename)


from typing import Any, Callable, Optional, Set

def cwd() -> None:
	from os.path import abspath
	from os import chdir
	from sys import path as sys_path
	''' Change the current directory to the base of relative paths to the directory
	and returns it. '''
	path = abspath(sys_path[0])
	chdir(path)
	return path

_load_formats = 'bytes', 'csv', 'json', 'jsonl', 'pickle', 'string', 'table', 'yaml'
def load(path: str, format: str = 'auto', encoding: str = 'utf-8', inner_args: list = None, inner_kwargs: dict = None) -> object:
	''' Load a file in a given format. '''
	format = detect_format(path, format, accept=_load_formats, default='string')
	args = [] if inner_args is None else inner_args
	kwargs = {} if inner_kwargs is None else inner_kwargs
	if format == 'string':
		fp = open(path, 'r', encoding=encoding)
	else:
		fp = open(path, 'rb')
	if format in ('bytes', 'string', 'jsonl'):
		res = fp.read()
	elif format == 'json':
		from json import load as jload
		res = jload(fp, *args, **kwargs)
	elif format == 'jsonl':
		from json import load as jload
		res = [jload(line, *args, **kwargs) for line in fp]
	elif format == 'csv':
		from pandas import read_csv
		res = read_csv(fp, *args, **kwargs)
	elif format == 'table':
		from pandas import read_excel
		res = read_excel(fp, *args, **kwargs)
	elif format == 'pickle':
		from pickle import load as pload
		res = pload(fp, *args, **kwargs)
	elif format == 'yaml':
		from yaml import safe_load_all as yload
		res = list(yload(fp, *args, **kwargs))
		res = res[0] if len(res) == 1 else res
	fp.close()
	return res

def save(path: str, content: object, format: str = 'auto', encoding: str = 'utf-8', append: bool = False, inner_args: list = None, inner_kwargs: dict = None) -> None:
	''' Saves a file to the given format. '''
	format = detect_format(path, format, accept=_load_formats, default='string')
	args = [] if inner_args is None else inner_args
	kwargs = {} if inner_kwargs is None else inner_kwargs
	if format in ('string', 'json', 'jsonl', 'yaml'):
		fp = open(path, 'a' if append else 'w', encoding=encoding)
	elif format in ('bytes', 'pickle'):
		fp = open(path, 'ab' if append else 'wb')
	else:
		fp = None
	if format in ('bytes', 'string'):
		fp.write(content)
	elif format in ('json', 'jsonl'):
		from json import dumps as jdumps
		if 'ensure_ascii' not in kwargs: kwargs['ensure_ascii'] = False
		if 'indent' not in kwargs: kwargs['indent'] = '\t'
		if format[-1] == 'l':
			fp.write(jdumps(elem, *args, **kwargs) for elem in content)
		else:
			fp.write(jdumps(content, *args, **kwargs))
	elif format == 'pickle':
		from pickle import dump as pdump
		if 'protocol' not in kwargs: kwargs['protocol'] = 4
		pdump(content, fp, *args, **kwargs)
	elif format == 'csv':
		from pandas import DataFrame
		if 'index' not in kwargs: kwargs['index'] = False
		if 'encoding' not in kwargs: kwargs['encoding'] = 'utf-8-sig'
		DataFrame(content).to_csv(path, *args, **kwargs)
	elif format == 'table':
		from pandas import DataFrame
		if 'index' not in kwargs: kwargs['index'] = False
		DataFrame(content).to_excel(path, *args, **kwargs)
	elif format == 'yaml':
		from yaml import dump as ydump
		ydump(content, fp, *args, **kwargs)
	if fp is not None:
		fp.close()

_decompress_formats = 'tar', 'zip', 'gzip', 'bzip2', 'rar', '7zip', 'lzma'
def decompress(input_file: str, output_dir: str = None, format: str = 'auto') -> None:
	''' Decompress the given file to the output directory regardless of its format. '''
	if output_dir is None:
		from os.path import dirname
		output_dir = dirname(input_file)
	else:
		from os import makedirs
		makedirs(output_dir, exist_ok=True)
	format = detect_format(input_file, format, accept=_decompress_formats)
	if format == 'zip':
		from zipfile import ZipFile
		with ZipFile(input_file, 'r') as i:
			i.extractall(output_dir)
	elif format == 'gzip':
		from gzip import GzipFile
		from os.path import join
		from shutil import copyfileobj
		o_dir = join(output_dir, input_file.rsplit('.', 1)[0])
		with GzipFile(input_file, 'r') as i, open(o_dir, 'wb') as o:
			copyfileobj(i, o)
	elif format == 'rar':
		from rarfile import RarFile
		with RarFile(input_file, 'r') as i:
			i.extractall(output_dir)
	elif format == 'bzip2':
		from bz2 import open as open_bz2
		from shutil import copyfileobj
		with open_bz2(input_file, 'r') as i, open(output_dir, 'wb') as o:
			copyfileobj(i, o)
	elif format == 'tar':
		from tarfile import open as open_tar
		with open_tar(input_file) as i:
			i.extractall(output_dir)
	elif format == '7zip':
		from py7zr import SevenZipFile
		with SevenZipFile(input_file, 'r') as i:
			i.extractall(output_dir)
	elif format == 'lzma':
		from lzma import open as open_lzma, decompress as lzdecompress
		from shutil import copyfileobj
		with open_lzma(input_file, 'r') as i, open(output_dir, 'wb') as o:
			copyfileobj(lzdecompress(i), o)

_detect_format_exts = (
	('bytes', ('bin', 'db', 'dat', 'blob', 'bytes')),
	('csv', ('csv',)),
	('json', ('json', 'js')),
	('jsonl', ('jsonl', 'jsl')),
	('pickle', ('pickle', 'pk', 'pkl', 'pck', 'pcl')),
	('string', ('txt', 'text', 'str')),
	('table', ('xlsx', 'odf', 'ods', 'odt', 'xls', 'xlsb', 'xlsm')),
	('yaml', ('yaml', 'yml')),
	('tar', ('tar', 'tar-linux32', 'tar-linux64', 'tar.gz')),
	('zip', ('zip', 'cbz')),
	('gzip', ('gz', 'gzip', 'gunzip')),
	('bzip2', ('bzip2', 'bz2')),
	('rar', ('rar', 'cbr')),
	('7zip', ('7z', '7zip')),
	('lzma', ('lzip', 'lz')),
)
def detect_format(path: str, format: str, accept: list = None, default: str = None) -> Optional[str]:
	''' Detects the format of a file from its path. '''
	if format == 'auto':
		name = path.lower()
		for ext_format, exts in _detect_format_exts:
			if any(name.endswith('.' + ext) for ext in exts):
				format = ext_format
				break
		else:
			format = default
	if accept is not None:
		assert format in accept, 'Unknown format for file at "%s". Accepted formats are %s.' % (
			path,
			', '.join(accept)
		)
	return format

def disk_cache(method=None, *, seconds: float = None, directory: str = '.cached/', identifier: str = None):
	''' The first time the decorated method is called, its result is stored as a pickle file, the
	next call loads the cached result from the disk. The cached files are used indefinitely unless the
	`seconds` lifespan is defined. The cached files are stored at `.cached` unless otherwise
	specificed with the `directory` argument. The cache file identifier is generated from the
	method name plus its arguments, unless otherwise specified with the `identifier` argument. '''
	def decorator(method):
		def wrapper(*args, **kwargs):
			from hashlib import md5
			from os import makedirs
			from os.path import join, exists
			from time import time
			makedirs(directory, exist_ok=True)
			# compute the cached file identifier
			if identifier is None:
				fname = '%s|%s|%s' % (
					method.__name__,
					','.join(map(str, args)),
					','.join(map(lambda it: '%s:%s' % it, kwargs.items()))
				)
				fname = md5(fname.encode()).hexdigest()
			else:
				fname = identifier
			fname = join(directory, str(fname))
			# return the value, cached or not
			res = None
			now = time()
			if exists(fname):
				from pickle import load as pload
				with open(fname, 'rb') as fp:
					save_time, value = pload(fp)
				if seconds is None or save_time - now < seconds:
					res = value
			if res is None:
				from pickle import dump as pdump
				res = method(*args, **kwargs)
				with open(fname, 'wb') as fp:
					pdump((now, res), fp)
			return res
		return wrapper
	if method:
		return decorator(method)
	else:
		return decorator

_mem_cache_global = {}
def mem_cache(
	method=None, *, key: Callable = None, maxsize: int = None, is_global: bool = False,
	global_name: str = None
):
	''' Decorator to cache the output of a method. It is indexed by its arguments
	unless the `key` argument is specified, in which case `key(*args, **kwargs)`
	will be called to get the indexing key. If `maxsize` is defined, it is bounded
	as an LRU cache with `maxsize` elements at most. If `is_global` is defined,
	the cache will be stored globally, so that it can be shared accross multiple
	methods of multiple instances of a class. A `global_name` can be
	defined to identify the method; otherwise, the method name will be used. '''
	if key is None:
		key = lambda *args, **kwargs: frozenset(args + tuple(kwargs.items()))
	if method is None:
		return lambda method: mem_cache(method, key=key, maxsize=maxsize, is_global=is_global, global_name=global_name)
	if is_global:
		if global_name is None: global_name = method.__name__
		cache, cache_usage = _mem_cache_global.get(global_name, ({}, []))
	else:
		cache, cache_usage = {}, []

	def _mem_cache(method, key, maxsize):
		from functools import wraps
		if maxsize is None:
			@wraps(method)
			def _mem_cache_wrapper(*args, **kwargs):
				k = key(*args, **kwargs)
				if k in cache: return cache[k]
				res = method(*args, **kwargs)
				cache[k] = res
				return res
		else:
			@wraps(method)
			def _mem_cache_wrapper(*args, **kwargs):
				k = key(*args, **kwargs)
				if k in cache:
					res = cache[k]
					cache_usage.remove(k)
				else:
					res = method(*args, **kwargs)
					cache[k] = res
					if len(cache) > maxsize:
						elem = cache_usage.pop(-1)
						del cache[elem]
				cache_usage.insert(0, k)
				return res
		return _mem_cache_wrapper
	return _mem_cache(method, key, maxsize)

def clear_global_mem_cache(global_name: str = None):
	''' Clears the global memory cache. '''
	if global_name is None:
		_mem_cache_global.clear()
	else:
		_mem_cache_global.pop(global_name, None)

def size(file) -> int:
	''' A way to see the size of a file without loading it to memory. '''
	if file.content_length:
		return file.content_length
	try:
		pos = file.tell()
		file.seek(0, 2)
		size = file.tell()
		file.seek(pos)
		return size
	except (AttributeError, IOError):
		pass
	return 0

_find_hidden_compressed_signatures = {
	'RNC': b'\x52\x4e\x43\x01',
	'RNC2': b'\x52\x4e\x43\x02',
	'lzip': b'\x4c\x5a\x49\x50',
	'zip': b'\x50\x4b\x03\x04',
	'zip-spanned': b'\x50\x4b\x07\x08',
	'rar1.5+': b'\x52\x61\x72\x21\x1a\x07\x00',
	'rar5.0+': b'\x52\x61\x72\x21\x1a\x07\x01\x00',
	'iso': b'\x43\x44\x30\x30\x31',
	'xar': b'\x78\x61\x72\x21',
	'tar1': b'\x75\x73\x74\x61\x72\x00\x30\x30',
	'tar2': b'\x75\x73\x74\x61\x72\x20\x20\x00',
	'7z': b'\x37\x7a\xbc\xaf\x27\x1c',
	'lz4': b'\x04\x22\x4d\x18',
	'webm': b'\x1a\x45\xdf\xa3',
	'xz': b'\xfd\x37\x7a\x58\x5a\x00',
	'wim': b'\x4d\x53\x57\x49\x4d\x00\x00',
	# signatures with too many false positives below
	# 'pdf': b'\x25\x50\x44\x46',
	# 'zip-empty': b'\x50\x4b\x05\x06',
	# 'gz': b'\x1f\x8b\x08',
	# 'tar': b'\x1f\x9d',
	# 'bz2': b'\x42\x5a\x68',
}
def find_hidden_compressed(path: str, byte_limit: int = None) -> Set[str]:
	''' Recursively looks for compressed file signatures in a file. '''
	with open(path, 'rb') as fp:
		data = fp.read() if byte_limit is None else fp.read(byte_limit)
	return {
		ftype
		for ftype, signature in _find_hidden_compressed_signatures.items()
		if data.find(signature) != -1
	}

_tvshow_rename_regex = None
def tvshow_rename(path: str) -> None:
	''' Rename every TV show of a folder.
	I.e. Inception_Season_4_Episode_02_DivX-Total.mkv would be 04x02.mkv. '''
	from os import listdir, rename
	global _tvshow_rename_regex
	if _tvshow_rename_regex is None:
		from regex import compile
		_tvshow_rename_regex = compile(r'(?P<SEASON>\d+)\s*[x\-]\s*(?P<EPISODE>\d+)|S\s*(?P<SEASON>\d+)\s*E\s*(?P<EPISODE>\d+)|(?P<EPISODE>\d+)').search
	for file in listdir(path):
		name, ext = file.rsplit('.', 1)
		match = _tvshow_rename_regex(name.replace('_', ' '))
		if match is not None:
			season, episode = match.groups()
			season = 1 if season is None else int(season)
			episode = int(episode)
			name = '%02d x %02d.%s' % (season, episode, ext)
			rename(file, name)

_directory_compare_ignored = ('.class', '.metadata', '.recommenders', '.pyc', '.git', '.svn', '.cached', '__pycache__')
def directory_compare(old: str, new: str, kind: str = 'dir', ignored: list = _directory_compare_ignored) -> None:
	''' Compares the files in two directories (old and new) to detect files that have been created,
	deleted, changed or updated, ignoring the specified files. '''
	def children(path):
		from os import listdir
		from os.path import isdir, islink, join
		res = {}
		for child in listdir(path):
			if not any(child.endswith(ext) for ext in ignored):
				full = join(path, child)
				is_dir = 'dir' if isdir(full) else 'file'
				if not islink(full):  # symbolic links are ignored in the comparison
					res[(is_dir, child)] = full
		return res
	if kind == 'file':
		from filecmp import cmp
		if not cmp(old, new, shallow=False):
			print('Modified\tfile\t%s' % new)
	else:
		old_childs, new_childs = children(old), children(new)
		for child in old_childs:
			if child not in new_childs:
				print('Deleted \t%s\t%s' % (child[0], old_childs[child]))
		for child in new_childs:
			if child not in old_childs:
				print('Created \t%s\t%s' % (child[0], new_childs[child]))
			else:
				directory_compare(old_childs[child], new_childs[child], child[0])

def import_from_path(path: str, name: str, module_name: str = '.') -> Any:
	''' Loads the script at the specified path and returns an object given its name. '''
	from importlib.util import spec_from_file_location, module_from_spec
	from os.path import exists
	if exists(path):
		spec = spec_from_file_location(module_name, path)
		module = module_from_spec(spec)
		spec.loader.exec_module(module)
		if name in module.__dict__:
			return module.__dict__[name]

def run_notebook(path: str) -> None:
	''' Runs a notebook and returns the result. '''
	source = ''
	for cell in load(path, 'json')['cells']:
		if cell['cell_type'] == 'code':
			source += ''.join(line for line in cell['source'] if not line.startswith('%')) + '\n'
	return exec(source, globals(), locals())


from typing import Dict, Optional, Tuple

def download_file(url, path=None, chunk_size=10**5, show_progress=True) -> str:
	''' Downloads a file keeping track of the progress. Returns the output path. '''
	from requests import get
	from simpler.format import human_seconds, human_bytes
	from sys import stdout
	from time import time
	if path is None: path = url.split('/')[-1]
	r = get(url, stream=True)
	total_bytes = int(r.headers.get('content-length'))
	bytes_downloaded = 0
	start = time()
	if show_progress: print('Downloading %s (%s)' % (url, human_bytes(total_bytes)))
	with open(path, 'wb') as fp:
		for chunk in r.iter_content(chunk_size=chunk_size):
			if not chunk: continue
			fp.write(chunk)
			if show_progress:
				bytes_downloaded += len(chunk)
				percent = bytes_downloaded / total_bytes
				bar = ('█' * int(percent * 32)).ljust(32)
				time_delta = time() - start
				eta = human_seconds((total_bytes - bytes_downloaded) * time_delta / bytes_downloaded)
				avg_speed = human_bytes(bytes_downloaded / time_delta).rjust(9)
				stdout.flush()
				stdout.write('\r  %6.02f%% |%s| %s/s eta %s' % (100 * percent, bar, avg_speed, eta))
	if show_progress: print()
	return path

class DownloaderPool:

	def __init__(self, num_workers=100, download_method=None):
		self.num_workers = num_workers
		self.pending_urls = []
		self.responses = {}
		self.workers = None
		if download_method is None:
			from urllib.request import urlopen
			self.download_method = lambda url: urlopen(url, timeout=5).read()
		else:
			self.download_method = download_method

	def spawn_workers(self):
		from threading import Thread
		self.workers = [Thread(target=self.download_worker) for _ in range(self.num_workers)]
		[w.start() for w in self.workers]

	def download_worker(self):
		from traceback import print_exc
		while len(self.pending_urls):
			url = self.pending_urls.pop()
			try:
				res = self.download_method(url)
			except:
				print_exc()
				res = None
			self.responses[url] = res

	def get(self, urls):
		self.pending_urls.extend(urls)
		request_pending_urls = urls[:]
		self.spawn_workers()
		while len(request_pending_urls):
			for url in request_pending_urls:
				if url in self.responses:
					break
			else:
				continue
			yield url, self.responses[url]
			del self.responses[url]
			request_pending_urls.remove(url)

_throttle_last = 0
def throttle(seconds: float = 1) -> None:
	''' Sleeps the thread so that the function is called every X seconds. '''
	global _throttle_last
	from time import sleep, time
	now = time()
	remaining = _throttle_last + seconds - now
	if remaining > 0:
		sleep(remaining)
		_throttle_last += seconds
	else:
		_throttle_last = now

class Driver:

	_CONSOLE_LEVELS = 'debug', 'info', 'log', 'warn', 'error'
	_YEAR_SECONDS = 365.4 * 24 * 3600
	_WAIT_POLL_EACH = .1

	def __init__(
		self, timeout: int = 3, keystroke_delay: int = .005, headless: bool = True, disable_flash: bool = True,
		disable_images: bool = True, language: str = 'en-US, en', options: Dict[str, str] = None
	):
		from autoselenium import Firefox
		from selenium.webdriver.firefox.options import Options
		opts = Options()
		opts.set_preference('intl.accept_languages', language)
		if options is not None:
			[opts.set_preference(k, v) for k, v in options.items()]
		self.driver = Firefox(headless=headless, disable_flash=disable_flash, disable_images=disable_images, options=opts)
		self.timeout = timeout
		self.keystroke_delay = keystroke_delay

	def browse(self, path: str):
		from selenium.webdriver.support.ui import WebDriverWait
		self.driver.get(path)
		WebDriverWait(self.driver, self.timeout).until(lambda d: d.execute_script('return document.readyState') == 'complete')
		self.driver.execute_script('window.CONSOLE_MESSAGES = [];%s' % ';'.join(
			'console.%s = (m) => CONSOLE_MESSAGES.push([\'%s\', m])' % (level, level)
			for level in self._CONSOLE_LEVELS
		))

	def wait(self, element: str, message: str = None, raise_errors: bool = True, invert: bool = False) -> bool:
		from selenium.common.exceptions import TimeoutException
		from selenium.webdriver.common.by import By
		from selenium.webdriver.support import expected_conditions as EC
		from selenium.webdriver.support.ui import WebDriverWait
		try:
			if invert:
				WebDriverWait(self.driver, self.timeout).until_not(EC.presence_of_element_located((By.CSS_SELECTOR, element)))
			else:
				WebDriverWait(self.driver, self.timeout).until(EC.presence_of_element_located((By.CSS_SELECTOR, element)))
			return True
		except TimeoutException as e:
			if raise_errors:
				if message is None:
					message = 'Timeout waiting for element to disappear: ' if invert else 'Timeout waiting for element to appear: '
				raise AssertionError(message + element) from e
		return False

	def wait_for_url(self, url: str, message: str = None, raise_errors: bool = True, invert: bool = False) -> bool:
		from selenium.common.exceptions import TimeoutException
		from selenium.webdriver.support import expected_conditions as EC
		from selenium.webdriver.support.ui import WebDriverWait
		try:
			if invert:
				WebDriverWait(self.driver, self.timeout).until_not(EC.url_to_be(url))
			else:
				WebDriverWait(self.driver, self.timeout).until(EC.url_to_be(url))
			return True
		except TimeoutException as e:
			if raise_errors:
				if message is None:
					message = 'Timeout waiting for URL to not be: ' if invert else 'Timeout waiting for URL to be: '
				raise AssertionError(message + url) from e
		return False

	def wait_for_file(self, path: str, message: str = 'Timeout waiting for file: ', raise_errors=True) -> bool:
		from os.path import exists
		from time import sleep, time
		start = time()
		while True:
			if exists(path): return True
			elapsed = time() - start
			if elapsed > self.timeout:
				if raise_errors:
					raise AssertionError(message + path)
				return False
			sleep(self._WAIT_POLL_EACH)

	def select(self, element, wait: bool = True, all: bool = False, raise_errors: bool = None):
		from selenium.webdriver.remote.webelement import WebElement
		if isinstance(element, WebElement): return element
		found = self.wait(element, raise_errors=not all if raise_errors is None else raise_errors) if wait else False
		if all:
			return self.driver.find_elements_by_css_selector(element)
		elif found:
			return self.driver.find_element_by_css_selector(element)

	def write(self, element, text: str, clear: bool = False) -> None:
		from selenium.webdriver.common.action_chains import ActionChains
		from selenium.webdriver.common.keys import Keys
		from time import sleep
		element = self.select(element)
		if clear:
			self.click(element)
			ActionChains(self.driver).key_down(Keys.CONTROL).send_keys('a').key_up(Keys.CONTROL).send_keys(Keys.BACKSPACE).perform()
		for char in text:
			element.send_keys(self.translate(char))
			sleep(self.keystroke_delay)

	def press(self, text: str):
		from selenium.webdriver.common.action_chains import ActionChains
		from time import sleep
		for char in text:
			ActionChains(self.driver).send_keys(self.translate(char)).perform()
			sleep(self.keystroke_delay)

	def click(self, element) -> None:
		self.select(element).click()

	def hover(self, element) -> None:
		from selenium.webdriver.common.action_chains import ActionChains
		ActionChains(self.driver).move_to_element(self.select(element)).perform()

	def focus(self, element) -> None:
		from selenium.webdriver.common.action_chains import ActionChains
		ActionChains(self.driver).move_to_element(self.select(element)).click().perform()

	def drag(self, element, x_offset: int = 0, y_offset: int = 0) -> None:
		from selenium.webdriver.common.action_chains import ActionChains
		ActionChains(self.driver).drag_and_drop_by_offset(self.select(element), x_offset, y_offset).perform()

	def scroll(self, element, x_delta: float = 0, y_delta: float = 0, mouse_x: float = 0, mouse_y: float = 0) -> None:
		self.driver.execute_script('''arguments[0].dispatchEvent(new WheelEvent("wheel", { bubbles: true,
			deltaX: arguments[1], deltaY: arguments[2], clientX: arguments[3], clientY: arguments[4]
		}))''', self.select(element), x_delta, y_delta, mouse_x, mouse_y)

	def scroll_into_view(self, element):
		self.driver.execute_script('arguments[0].scrollIntoView({"block":"center"});', self.select(element))

	def attribute(self, element, attribute: str, value: Optional[str] = None) -> Optional[str]:
		args = [self.select(element), attribute]
		if value is None:
			script = 'return !arguments[0].hasAttribute(arguments[1]) ? null : arguments[0].getAttribute(arguments[1])'
		else:
			script = 'arguments[0].setAttribute(arguments[1], arguments[2])'
			args.append(value)
		return self.driver.execute_script(script, *args)

	def style(self, element, property: str, value: Optional[str] = None) -> Optional[str]:
		args = [self.select(element), property]
		if value is None:
			script = 'return getComputedStyle(arguments[0])[arguments[1]]'
		else:
			script = 'arguments[0].style[arguments[1]] = arguments[2]'
			args.append(value)
		return self.driver.execute_script(script, *args)

	def has_class(self, element, class_name: str) -> bool:
		return self.driver.execute_script(
			'return arguments[0].classList.contains(arguments[1])',
			self.select(element),
			class_name
		)

	def cookie(
		self, name: str, value: Optional[str] = None, expiry: int = None, delete: bool = False,
		path: str = None, domain: str = None, http_only: str = None, secure: str = None
	) -> Optional[str]:
		from time import time
		if delete:
			self.driver.delete_cookie(name)
		elif value is None:
			return self.driver.get_cookie(name)['value']
		else:
			cookie = {'name': name, 'value': value, 'path': '/'}
			cookie['expiry'] = expiry if expiry is not None else int(time() + self._YEAR_SECONDS)
			if path is not None: cookie['path'] = path
			if domain is not None: cookie['domain'] = domain
			if http_only is not None: cookie['httpOnly'] = http_only
			if secure is not None: cookie['secure'] = secure
			self.driver.add_cookie(cookie)

	def local_storage(self, key: str, value: Optional[str] = None, delete: bool = False) -> Optional[str]:
		if delete:
			self.driver.execute_script('localStorage.removeItem(arguments[0])', key)
		elif value is None:
			return self.driver.execute_script('return localStorage.getItem(arguments[0])', key)
		else:
			self.driver.execute_script('localStorage.setItem(arguments[0], arguments[1])', key, value)

	def session_storage(self, key: str, value: Optional[str] = None, delete: bool = False) -> Optional[str]:
		if delete:
			self.driver.execute_script('sessionStorage.removeItem(arguments[0])', key)
		elif value is None:
			return self.driver.execute_script('return sessionStorage.getItem(arguments[0])', key)
		else:
			self.driver.execute_script('sessionStorage.setItem(arguments[0], arguments[1])', key, value)

	def all_cookies(self, clear: bool = True, path: str = None, domain: str = None, http_only: str = None, secure: str = None) -> dict:
		res = {
			c['name']: c['value']
			for c in self.driver.get_cookies()
			if (path is None or c['path'] == path)
			and (domain is None or c['domain'] == domain)
			and (http_only is None or c['httpOnly'] == http_only)
			and (secure is None or c['secure'] == secure)
		}
		if clear:
			self.driver.delete_all_cookies()
		return res

	def all_local_storage(self, clear: bool = True) -> dict:
		res = {k: v for k, v in self.driver.execute_script('return Object.entries(localStorage)')}
		if clear:
			self.driver.execute_script('localStorage.clear()')
		return res

	def all_session_storage(self, clear: bool = True) -> dict:
		res = {k: v for k, v in self.driver.execute_script('return Object.entries(sessionStorage)')}
		if clear:
			self.driver.execute_script('sessionStorage.clear()')
		return res

	def box(self, element) -> Tuple[float, float]:
		res = self.driver.execute_script('return arguments[0].getBoundingClientRect()', self.select(element))
		res['center_left'] = res['left'] + res['width'] / 2
		res['center_top'] = res['top'] + res['height'] / 2
		return res

	def translate(self, char: str) -> str:
		from selenium.webdriver.common.keys import Keys
		if char == '\t':
			return Keys.TAB
		elif char == '\n':
			return Keys.ENTER
		else:
			return char

	def console_clear(self):
		self.driver.execute_script('window.CONSOLE_MESSAGES = []')

	def console_messages(self, group_by_level: bool = False) -> Dict[str, str]:
		messages = self.driver.execute_script('return CONSOLE_MESSAGES')
		if group_by_level:
			res = {}
			for level, message in messages:
				res.setdefault(level, []).append(message)
			return res
		else:
			return [m[1] for m in messages]


def compose(from_mail: str, dest_mail: object, from_name: str = None, dest_name: object = None, text: str = '', text_type: str = 'plain', subject: str = '', blocking: bool = False):
	from email.message import EmailMessage
	from email.headerregistry import Address
	from email.policy import default
	message = EmailMessage(policy=default)
	message['From'] = Address(from_name, from_mail)
	message['To'] = Address(dest_name, dest_mail)
	message['Subject'] = subject
	if text_type == 'plain':
		message.set_content(text)
	else:
		message.set_content(text, subtype='html')
	return message

def send(smtp_server: str, password: str, mail: str, message) -> None:
	''' Sends a message. '''
	from email.message import EmailMessage
	from smtplib import SMTP
	assert isinstance(message, EmailMessage), 'Message must be an instance of email.message.EmailMessage; you can use simpler.compose to generate it.'
	server = SMTP(smtp_server)
	server.starttls()
	server.login(mail, password)
	server.send_message(message)
	server.quit()


_tictoc_stack = []
def tic() -> None:
	''' Captures time '''
	from time import time
	_tictoc_stack.append(time())

def toc(show: bool = True, show_label: str = '') -> float:
	''' Shows time since tic() was executed. '''
	from time import time
	total = time() - _tictoc_stack.pop()
	if show:
		if show_label:
			print(show_label, end=': ')
		print(total)
	return total

_deep_size_blacklist = None
def deep_size(obj):
	''' Get the actual size of an instance, exploring all its references. '''
	global _deep_size_blacklist
	from sys import getsizeof
	from types import ModuleType, FunctionType
	from gc import get_referents
	if _deep_size_blacklist is None:
		_deep_size_blacklist = type, ModuleType, FunctionType
	if isinstance(obj, _deep_size_blacklist):
		raise TypeError('getsize() does not take argument of type: ' + str(type(obj)))
	seen_ids = set()
	size = 0
	objects = [obj]
	while objects:
		need_referents = []
		for obj in objects:
			if not isinstance(obj, _deep_size_blacklist) and id(obj) not in seen_ids:
				seen_ids.add(id(obj))
				size += getsizeof(obj)
				need_referents.append(obj)
		objects = get_referents(*need_referents)
	return size


from simpler._version import __version__
from simpler.algorithms import DynamicProgramming
from simpler.bioinformatics import codon_table, monoisotopic_mass_table, monoisotopic_mass_water, parse_fasta, dna_to_rna, rna_to_dna, rna_to_protein, reverse_complement
from simpler.connectors import SQL, Excel
from simpler.files import cwd, load, save, disk_cache, mem_cache, clear_global_mem_cache, size, find_hidden_compressed, tvshow_rename, directory_compare, decompress, import_from_path
from simpler.format import human_bytes, human_seconds, human_date, random_string, print_matrix, safe_filename
from simpler.mail import compose, send
from simpler.math import clamp, snap, unique, all_equal, jaccard, levenshtein, base_change, prime_list, is_prime, fibonacci, lcm, gcd, factor, palindrome_list, phi, date_range
from simpler.profiling import tic, toc, deep_size
from simpler.sparql import dbpedia, entity_types
from simpler.terminal import getch, cprint
from simpler.tests import Test, Suite
from simpler.validation import assert_set, assert_str, assert_number, assert_id, assert_mail, assert_exists
from simpler.web import download_file, DownloaderPool, throttle, Driver


class DynamicProgramming:
	'''
	Abstract class to solve problems using dynamic programming. To use it,
	make a child class implementing alternatives, is_final and penalty. Then,
	make one of such objects providing the initial state in the constructor, and
	call to solve(), providing one search type.
	'''

	ONE_SOLUTION = 0
	ONE_OPTIMAL_SOLUTION = 1
	ALL_SOLUTIONS = 2
	ALL_OPTIMAL_SOLUTIONS = 3

	def __init__(self, initial_state):
		self.initial_state = initial_state

	def alternatives(self, state: object) -> list:
		''' This should return all the alternatives for a given state, without modifying the given state. '''
		raise NotImplementedError()

	def is_final(self, state: object) -> bool:
		''' This should return whether a state is a final state or not. '''
		raise NotImplementedError()

	def penalty(self, state: object) -> float:
		''' This should return an upper bound for the penalty of the problem. Ideally, optimal solutions should have no penalty. '''
		raise NotImplementedError()

	def solve(self, search_type: int = ONE_SOLUTION):
		''' Resolves the Dynamic Programming problem. '''
		assert search_type in range(4), "Invalid search type."
		final_states = []
		explored = []
		remaining = [(self.penalty(self.initial_state), self.initial_state)]
		while len(remaining):
			_, state = remaining.pop(0)
			explored.append(state)
			for alternative in self.alternatives(state):
				if alternative in explored: continue
				penalty = self.penalty(alternative)
				if self.is_final(alternative):
					if search_type == self.ONE_SOLUTION or penalty == 0 and search_type == self.ONE_OPTIMAL_SOLUTION:
						return alternative
					else:
						final_states.append((penalty, alternative))
				else:
					index = sum(1 for p, _ in remaining if p < penalty)
					remaining.insert(index, (penalty, alternative))
		if search_type == self.ALL_SOLUTIONS:
			return [state for _, state in final_states]
		else:
			min_penalty = min([penalty for penalty, _ in final_states])
			optimals = [state for penalty, state in final_states if penalty == min_penalty]
			if search_type == self.ALL_OPTIMAL_SOLUTIONS:
				return optimals
			else:
				return optimals[0]


from typing import Dict, List, Optional, Tuple

class Test:
	''' Class for a test case. Each test case might contain multiple `test_<something>` methods
	which are called when running the tests. It is advisable to run it from the `simpler.Suite` class. '''
	PREFIX = 'test_'

	def run(self) -> Dict[str, Tuple[Optional[str], float]]:
		''' Internal method used to run the tests. '''
		from time import time
		from traceback import format_exc
		res = {}
		total_elapsed = passes = errors = 0
		for name in self.__dir__():
			if not name.startswith(self.PREFIX): continue
			method = self.__getattribute__(name)
			elapsed = time()
			try:
				method()
				output = None
			except AssertionError as e:
				output = e.args[0]
			except:
				output = 'Error while running the test:\n' + format_exc()
			elapsed = time() - elapsed
			total_elapsed += elapsed
			res[name[len(self.PREFIX):]] = output, elapsed
			if output is None:
				passes += 1
			else:
				errors += 1
		return res, total_elapsed, passes, errors

class Suite:
	''' Class for running a test suite. Is built as Suite(FirstTest, SecondTest...) where the
	arguments are an enumeration of subclasses of `simpler.Test` classes that will be run when using
	this class `run` method. There are a few `run_<format>` methods to get a formatted output. '''

	def __init__(self, *tests: List[Test]) -> None:
		assert all(t.__name__.endswith('Test') for t in tests), 'All test classes must end with "Test".'
		self.tests = tests

	def run(self) -> Tuple[Dict[str, Tuple[Dict[str, Tuple[Optional[str], float]], float, int, int]], float, int, int]:
		''' Runs the tests and returns a dictionary of tests, where each test is a dictionary of
		`case: error` pairs. '''
		res = {}
		total_elapsed = total_errors = total_passes = 0
		for test in self.tests:
			output, elapsed, passes, errors = test().run()
			total_elapsed += elapsed
			total_errors += errors
			total_passes += passes
			if len(output):
				res[test.__name__[:-4]] = output, elapsed, errors, passes
		return res, total_elapsed, total_passes, total_errors

	def run_text(self, only_errors: bool = True) -> Optional[str]:
		''' Runs the `run` method and formats the output as a plain text. '''
		tests, elapsed, passes, errors = self.run()
		if errors or not only_errors:
			res = 'Summary: %d errors and %d passes in %.2fs.\n' % (errors, passes, elapsed)
			for test, (cases, test_elapsed, test_passes, test_errors) in tests.items():
				if only_errors and test_errors == 0: continue
				res += 'Test %s (%d/%d in %.2fs)\n' % (test, test_errors, test_errors + test_passes, test_elapsed)
				for case, (message, elapsed) in cases.items():
					if only_errors and message is None: continue
					res += '\tCase %s (%.2fs)' % (case, elapsed)
					res += ': PASS\n' if message is None else ': ERROR\n%s\n' % '\n'.join('\t\t%s' % line for line in message.strip().split('\n'))
			return res.strip()

	def run_html(self, only_errors: bool = True) -> Optional[str]:
		''' Runs the `run` method and returns a table of errors, or None if there isn't any. '''
		tests, elapsed, passes, errors = self.run()
		if errors or not only_errors:
			res = '<p><b>Summary: %d errors and %d passes in %.2fs.</b></p><table><tr><th>Test</th><th>Case</th><th>Elapsed (s)</th><th>%s</th></tr>' % (
				errors, passes, elapsed, 'Error' if only_errors else 'Result'
			)
			for test, (cases, _, _, _) in tests.items():
				for case, (message, elapsed) in cases.items():
					if message is not None or not only_errors:
						res += '<tr><td>%s</td><td>%s</td><td>%.2f</td><td><div style="font-family:monospace;white-space:pre;max-width:35rem;overflow:auto">%s</div></td></tr>' % (test, case, elapsed, ('<span style="color:#080">✓</span>' if message is None else message))
			if not only_errors:
				res += '<tr><th colspan=2>Total</th><td>%.2f</td><td>%d/%d OK</td></tr>' % (elapsed, passes, passes + errors)
			return res + '</table>'


from typing import List, Tuple

DBPEDIA_PREFIX = 'PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>\nPREFIX : <http://dbpedia.org/resource/>\nPREFIX dbc: <http://dbpedia.org/resource/Category:>\nPREFIX dct: <http://purl.org/dc/terms/>\nPREFIX dbo: <http://dbpedia.org/ontology/>\n\n'
QUERY_CLASSES = '''select ?type count (*) as ?num where {
    quad map virtrdf:DefaultQuadMap {
        graph ?g {
            ?s1 ?s1textp ?o1 .
            ?o1 bif:contains  '"%s"'
        }
    }
    ?s1 a ?type .
} group by ?type order by desc 2 limit 100 offset 0'''

def dbpedia(query: str):
    ''' Sends a query to DBPedia and return the results. '''
    from SPARQLWrapper import SPARQLWrapper, JSON
    res = SPARQLWrapper("http://dbpedia.org/sparql")
    res.setReturnFormat(JSON)
    res.setQuery(DBPEDIA_PREFIX + query)

    return res.query().convert()['results']['bindings']

def entity_types(value) -> List[Tuple[str, int]]:
    ''' Return every entity type with values that contain a given string sorted by frequency.'''
    res = dbpedia(QUERY_CLASSES % value.replace(' ', ' AND '))
    return [(r['type']['value'], int(r['num']['value'])) for r in res]